﻿Option Strict On
Imports Microsoft.VisualBasic
Imports System.Xml
Imports System.Security.Cryptography
Imports System.Security.Cryptography.Xml

#Const TRACE = True

Public Enum ConcurrencyResults
    Success = 0
    LicenseCountExceeded = 1
    LicenseFreedOrRevoked = 2
    CertificateInvalid = 3
    ServerCryptographyError = 4
    AppNotSupported = 5
    InvalidKeyForApplication = 6
    CertConcurrencyDataError = 7
    FunctionDisabled = 8

End Enum

Public Class InstallKeyInfo
    Private m_AllocatedLicenseList As New Dictionary(Of String, Date)
    Private m_MaxCount As Integer = 2
    Private m_LicenseLifetime As Integer = 300 ' Lifetime of each license in seconds
    Private m_KeyInfoCertDate As Date ' Date of certificate that sent MaxCount
    Private m_LastAccessTime As Date ' Date/Time of most recent access to this object

    Public Sub SetKeyInfo(ByVal count As Integer, ByVal lifetime As Integer, ByVal certdate As Date)
        m_MaxCount = count
        m_LicenseLifetime = lifetime
        m_KeyInfoCertDate = certdate
        m_LastAccessTime = Now
    End Sub

    Public Sub New(ByVal count As Integer, ByVal lifetime As Integer, ByVal certdate As Date)
        SetKeyInfo(count, lifetime, certdate)
        m_LastAccessTime = Now
    End Sub

    Public Sub New()
        MyBase.New()
    End Sub

    Public ReadOnly Property KeyInfoCertDate() As Date
        Get
            Return m_KeyInfoCertDate
        End Get
    End Property

    Public ReadOnly Property LicenseAvailable() As Boolean
        Get
            ClearOldLicenses()
            Return m_AllocatedLicenseList.Count < m_MaxCount
        End Get
    End Property

    Friend ReadOnly Property LastAccessTime() As Date
        Get
            Return m_LastAccessTime
        End Get
    End Property

    Public Sub AddOrUpdateUnique(ByVal UniqueInstall As String)
        m_LastAccessTime = Now
        m_AllocatedLicenseList.Item(UniqueInstall) = Now
    End Sub

    Public Function Renew(ByVal UniqueInstall As String) As ConcurrencyResults
        m_LastAccessTime = Now
        If m_AllocatedLicenseList.ContainsKey(UniqueInstall) Then
            m_AllocatedLicenseList.Item(UniqueInstall) = Now
            Return ConcurrencyResults.Success
        Else
            Return ConcurrencyResults.LicenseFreedOrRevoked
        End If
    End Function

    Public Sub RemoveUnique(ByVal UniqueInstall As String)
        m_LastAccessTime = Now
        m_AllocatedLicenseList.Remove(UniqueInstall)
    End Sub

    Private Sub ClearOldLicenses()
        Dim current As Integer = 0
        Do Until current >= m_AllocatedLicenseList.Count
            If Now.Subtract(m_AllocatedLicenseList.ElementAt(current).Value).TotalSeconds > m_LicenseLifetime Then
                ConcurrencyManager.Tracer.TraceInformation("Expiring Unique " & m_AllocatedLicenseList.ElementAt(current).Key)
                m_AllocatedLicenseList.Remove(m_AllocatedLicenseList.ElementAt(current).Key)
            Else
                current += 1
            End If
        Loop
    End Sub

End Class

Public Class ConcurrencyManager

    ' The m_AllocatedLicense list includes a list of installation keys
    ' The entry for each key is a list of unique installs of allocated licenses
    ' Each allocated license has a date - when it was last renewed
    Private m_AllocatedLicense As New Dictionary(Of String, InstallKeyInfo)
    Public Shared Tracer As New Diagnostics.TraceSource("ConcurrencyManagerTrace")

    Friend Function Allocate(ByVal Installkey As String, ByVal UniqueInstall As String) As ConcurrencyResults
        ' Grab the list for the existing key
        Dim keyinfo As InstallKeyInfo = Nothing
        SyncLock m_AllocatedLicense
            If m_AllocatedLicense.TryGetValue(Installkey, keyinfo) Then
                ' Install key already exists
                If Not keyinfo.LicenseAvailable Then
                    Tracer.TraceInformation("License count exceeded key=" & Installkey & " unique=" & UniqueInstall)
                    Return ConcurrencyResults.LicenseCountExceeded
                End If
                keyinfo.AddOrUpdateUnique(UniqueInstall)
                Tracer.TraceInformation("License granted key=" & Installkey & " unique=" & UniqueInstall)
            Else
                ' It's a new one - up to the caller to reset default values
                keyinfo = New InstallKeyInfo()
                keyinfo.AddOrUpdateUnique(UniqueInstall)
                m_AllocatedLicense.Add(Installkey, keyinfo)
                Tracer.TraceInformation("License granted key=" & Installkey & " unique=" & UniqueInstall)
            End If
        End SyncLock

        Return ConcurrencyResults.Success


    End Function

    Public Function Renew(ByVal InstallKey As String, ByVal UniqueInstall As String) As ConcurrencyResults
        ' Grab the list for the existing key
        Dim keyinfo As InstallKeyInfo = Nothing
        ' If license can't be renewed, it means it was timed out, or the application cycled
        ' This is a signal from the client app to reacquire a license
        SyncLock m_AllocatedLicense
            If Not m_AllocatedLicense.TryGetValue(InstallKey, keyinfo) Then
                Tracer.TraceInformation("License revoked key=" & InstallKey & " unique=" & UniqueInstall)
                Return ConcurrencyResults.LicenseFreedOrRevoked
            End If
            Dim results As ConcurrencyResults
            results = keyinfo.Renew(UniqueInstall)
            Tracer.TraceInformation("License renewed key=" & InstallKey & " unique=" & UniqueInstall & " status: " & results.ToString)
            Return results
        End SyncLock

    End Function

    Public Function Release(ByVal InstallKey As String, ByVal UniqueInstall As String) As ConcurrencyResults
        ' Grab the list for the existing key
        Dim keyinfo As InstallKeyInfo = Nothing
        ' If license can't be renewed, it means it was timed out, or the application cycled
        ' This is a signal from the client app to reacquire a license
        SyncLock m_AllocatedLicense
            If Not m_AllocatedLicense.TryGetValue(InstallKey, keyinfo) Then
                Tracer.TraceInformation("License revoked key=" & InstallKey & " unique=" & UniqueInstall)
                Return ConcurrencyResults.LicenseFreedOrRevoked
            End If
            Tracer.TraceInformation("License released key=" & InstallKey & " unique=" & UniqueInstall)
            keyinfo.RemoveUnique(UniqueInstall)
        End SyncLock
        Return ConcurrencyResults.LicenseFreedOrRevoked
    End Function

    ' Sets concurrency info for an installation key
    ' Only updates if info is newer
    Friend Sub SetConcurrencyInfo(ByVal Installkey As String, ByVal MaxCount As Integer, ByVal LicenseTime As Integer, ByVal InfoDate As Date)
        Dim keyinfo As InstallKeyInfo = Nothing
        If m_AllocatedLicense.TryGetValue(Installkey, keyinfo) Then
            ' Found one
            If InfoDate > keyinfo.KeyInfoCertDate Then
                ' This cert has a later date than the existing one
                Tracer.TraceInformation("Updating concurrency info. MaxCount=" & MaxCount.ToString & " LicenseTime=" & LicenseTime.ToString & " InfoDate=" & InfoDate.ToShortDateString)
                keyinfo.SetKeyInfo(MaxCount, LicenseTime, InfoDate)
             End If
        Else
            ' It's a new one
            Tracer.TraceInformation("Setting new concurrency info. MaxCount=" & MaxCount.ToString & " LicenseTime=" & LicenseTime.ToString & " InfoDate=" & InfoDate.ToShortDateString)
            keyinfo = New InstallKeyInfo(MaxCount, LicenseTime, InfoDate)
            m_AllocatedLicense.Add(Installkey, keyinfo)
        End If
        FlushUnusedCachedKeys()
    End Sub

    Friend Sub FlushUnusedCachedKeys()
        Static expiredminutes As Integer = CInt(ConfigurationManager.AppSettings.Item("UnusedInstallKeyHoldTimeMinutes"))
        Dim current As Integer = 0
        Do Until current >= m_AllocatedLicense.Count
            If Now.Subtract(m_AllocatedLicense.ElementAt(current).Value.LastAccessTime).TotalMinutes > expiredminutes Then
                Tracer.TraceInformation("Flushing unused key: " & m_AllocatedLicense.ElementAt(current).Key)
                m_AllocatedLicense.Remove(m_AllocatedLicense.ElementAt(current).Key)
            Else
                current += 1
            End If
        Loop
    End Sub

    Friend Function ParseCertificate(ByVal certificate As String, ByVal LicenseLifetime As Integer, ByRef InstallCode As String, ByRef UniqueCode As String) As ConcurrencyResults
        Dim xdoc As New XmlDocument
        Try
            xdoc.PreserveWhitespace = False
            xdoc.LoadXml(certificate)
        Catch ex As Exception
            Return ConcurrencyResults.CertificateInvalid
        End Try
        ' Now validate the certificate
        ' Find the application name
        Dim appname As String
        Try
            appname = xdoc.SelectSingleNode("//Application/Name").FirstChild.Value
        Catch ex As Exception
            Return ConcurrencyResults.CertificateInvalid
        End Try

        ' Get the key
        Dim pkxml As String
        Try
            pkxml = ConfigurationManager.AppSettings.Item(appname & "-pk")
        Catch ex As Exception
            Return ConcurrencyResults.AppNotSupported
        End Try

        Dim rsakey As RSACryptoServiceProvider
        Dim params As New CspParameters
        params.Flags = CspProviderFlags.NoPrompt
        Try
            rsakey = New RSACryptoServiceProvider(1024, params)
        Catch ex As Exception
            Try
                params.Flags = params.Flags Or CspProviderFlags.UseMachineKeyStore
                rsakey = New RSACryptoServiceProvider(1024, params)
            Catch ex2 As Exception
                Return ConcurrencyResults.ServerCryptographyError
            End Try

        End Try

        ' Load the key 
        Try
            rsakey.FromXmlString(pkxml)
        Catch ex As Exception
            rsakey.Clear()
            Return ConcurrencyResults.InvalidKeyForApplication
        End Try

        ' Now validate
        Dim XmlSignHolder As SignedXml
        Dim sig As XmlElement

        Try
            sig = CType(xdoc.GetElementsByTagName("Signature").Item(0), XmlElement)
            If sig Is Nothing Then Return ConcurrencyResults.CertificateInvalid
            XmlSignHolder = New SignedXml(xdoc)
            XmlSignHolder.LoadXml(sig)

            If Not XmlSignHolder.CheckSignature(rsakey) Then
                Return ConcurrencyResults.CertificateInvalid
            End If
        Catch ex2 As CryptographicException
            Return ConcurrencyResults.CertificateInvalid
        Finally
            rsakey.Clear()
        End Try

        ' Now pull the hashtable
        Try
            Dim sv As String = xdoc.GetElementsByTagName("ServerData").Item(0).FirstChild.Value
            Dim b As Byte() = Convert.FromBase64String(sv)
            Dim ms As New IO.MemoryStream(b)
            Dim serverdata As IDictionary = Desaware.DLSextensions.DictionarySerializer.Deserialize(ms)
            InstallCode = CStr(serverdata.Item("InstallationCode"))
            SetConcurrencyInfo(InstallCode, CInt(serverdata.Item("ConcurrencyCount")), LicenseLifetime, CDate(serverdata.Item("ConcurrencyDate")))
            UniqueCode = xdoc.SelectSingleNode("//UniqueInstallation/GUID").FirstChild.Value
        Catch ex As Exception
            Return ConcurrencyResults.CertConcurrencyDataError
        End Try



    End Function

End Class